import argparse
import sys
from getpass import getpass

from gtagoraapp.details.app import App
from gtagoraapp.details.settings import Settings

def get_input(text, existing_value=None, has_default=False):
    if existing_value:
        return f'{text} [ENTER={existing_value}]: '
    elif has_default:
        return f'{text} [ENTER=default]: '
    else:
        return f'{text}: '


def query_yes_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('--setup', action='store_true')
    args = parser.parse_args()

    settings = Settings()
    if args.setup or not settings.is_complete():
        server = input(get_input('agora server', settings.server))
        server = server.strip() if server else settings.server
        username = input(get_input('username'))
        password = getpass()
        download_path = input(get_input('download path', settings.download_path))
        download_path = download_path.strip() if download_path else settings.download_path
        logfile = input(get_input('logfile path', settings.logfile, True))
        logfile = logfile.strip() if logfile else settings.logfile
        log_level = input(get_input('log level', settings.log_level, True))
        log_level = log_level.upper() if log_level else settings.log_level
        console_log_level = None
        if query_yes_no('do console logging?'):
            console_log_level = input(get_input('console log level', settings.console_log_level, True))
            console_log_level = console_log_level.upper() if console_log_level else settings.console_log_level

        session_key = App.get_session_key(server, username, password)
        settings.setup(server, download_path, session_key, logfile, log_level, console_log_level)
        if args.setup:
            exit(0)

    app = App()
    app.run()



import asyncio
import json
import os
import platform
import socket
import _thread
from pathlib import Path
from urllib.parse import urlparse

import pkg_resources
import websockets
from websockets.http import Headers

from gtagoraapp.details.settings import Settings

class AgoraWebsocketMessage:
    def __init__(self, data: dict):
        self.msg = dict()
        self.msg['stream'] = 'App'
        self.msg['data'] = data

    def __str__(self):
        return json.dumps(self.msg)


class AgoraWebsocket:
    def __init__(self, settings: Settings, handler):
        self.settings = settings
        self.handler = handler
        self.logger = handler.logger

        uri = urlparse(self.settings.server)
        if not uri.hostname:
            raise ValueError('The server URL is invalid. Please run "gtagora --setup" to modify it')

        self.uri = 'ws://' + uri.hostname
        if uri.port:
            self.uri = f'{self.uri}:{uri.port}'

        self.ping_data = self._get_ping_data()

        self.ping_timeout = 10
        self.timout = 10
        self.sleep_time = 30

    def run(self):
        asyncio.get_event_loop().run_until_complete(self._listen_forever())

    async def _listen_forever(self):
        while True:
            # outer loop restarted every time the connection fails
            try:
                headers = Headers()
                headers['Authorization'] = f'Token {self.settings.session_key}'
                async with websockets.connect(self.uri, extra_headers=headers) as ws:
                    self.logger.info(f'Websocket established with {self.uri}')
                    self.logger.debug(f'Ping response: {json.dumps(self.ping_data)}')
                    await ws.send(str(AgoraWebsocketMessage(self.ping_data)))
                    self.logger.info(f'App is running --> listening for server messages...')
                    while True:
                        # listener loop
                        try:
                            msg = await ws.recv()
                        except (asyncio.TimeoutError, websockets.exceptions.ConnectionClosed) as e:
                            try:
                                self.logger.warning(f'Connection probably closed (sending ping): {str(e)}')
                                pong = await ws.ping()
                                await asyncio.wait_for(pong, timeout=self.ping_timeout)
                                continue
                            except:
                                self.logger.warning(f'Connection closed (trying again in {self.sleep_time}s): {str(e)}')
                                await asyncio.sleep(self.sleep_time)
                                break  # inner loop
                        asyncio.create_task(self._process_message(msg, ws))
            except socket.gaierror as e:
                self.logger.warning(f'Connection closed: {str(e)}')
                self.logger.warning(f'Trying to reconnect')
                continue
            except ConnectionRefusedError as e:
                self.logger.warning(f'Connection refused: {str(e)}')
                self.logger.warning(f'Trying to reconnect')
                continue

    async def _process_message(self, msg_str, ws):
        try:
            msg = json.loads(msg_str)
            data = msg.get('data')
            stream = msg.get('stream')
            if stream == 'App' and data and self._i_am_receiver(msg):
                command = data.get('command')
                if command == 'hello':
                    self.logger.info(f'Received Hello --> sending ping response')
                    self.logger.debug(f'Ping response: {json.dumps(self.ping_data)}')
                    await ws.send(str(AgoraWebsocketMessage(self.ping_data)))
                if command == 'download' and 'data' in data:
                    download_data = data['data']
                    _thread.start_new_thread(self.handler.download, (download_data.get('files'),))
                if command == 'runTask' and 'data' in data:
                    task_data = data['data']
                    _thread.start_new_thread(self.handler.run_task, (task_data,))
        except:
            pass

    def _i_am_receiver(self, msg):
        receiver = msg.get('receiver')
        if not receiver or receiver == self.settings.app_id:
            return True

        return False

    def _get_ping_data(self):
        system = 'unknown'
        operating_system = platform.system()
        if operating_system == 'Linux' or operating_system == 'Darwin':
            system = 'unix'
        elif operating_system == 'Windows':
            system = 'windows'

        try:
            version_str = pkg_resources.require("gtagora-app")[0].version
        except:
            version_str = '0.0.1'
        self.logger.debug(f'gtagora-app version: {version_str}')
        version = self.parse_version(version_str)
        self.logger.debug(f'gtagora-app version (parsed): {json.dumps(version)}')

        command_data = dict()
        command_data['appId'] = self.settings.app_id
        command_data['base_path'] = Path(self.settings.download_path).as_posix()
        command_data['computerName'] = platform.node()
        command_data['path_separator'] = os.path.sep
        command_data['system'] = system
        command_data['version'] = version

        data = dict()
        data['command'] = 'ping'
        data['data'] = command_data

        return data

    @staticmethod
    def parse_version(version_str: str):
        version = {'major': 0, 'minor': 0, 'path': 0, 'snapshot': False, 'string': version_str}

        version_str = version_str.strip()
        if 'SNAPSHOT' in version_str.upper():
            version['snapshot'] = True

        version_str = version_str.split('-')[0]
        version_str = version_str.split(' ')[0]
        version_splitted = version_str.split('.')
        counter = 0
        for part in version_splitted:
            if counter == 0:
                try:
                    version['major'] = int(part)
                except:
                    pass
            elif counter == 1:
                try:
                    version['minor'] = int(part)
                except:
                    pass
            elif counter == 2:
                try:
                    version['path'] = int(part)
                except:
                    pass
            counter += 1

        return version



import json
import logging
import os
from uuid import uuid4

from pathlib import PurePath, Path


class Settings:
    def __init__(self):
        json = self._get()
        self.server = json['server'] if 'server' in json else None
        self.download_path = json['download_path'] if 'download_path' in json else None
        self.session_key = json['session_key'] if 'session_key' in json else None
        self.app_id = json['app_id'] if 'app_id' in json else None
        self.logfile = json['logfile'] if 'logfile' in json else None
        self.log_level = json['log_level'] if 'log_level' in json else None
        self.console_log_level = json['console_log_level'] if 'console_log_level' in json else None

    def save(self):
        settings_file = self._get_or_create_file()
        with open(settings_file, 'w') as json_file:
            return json.dump(self.__dict__, json_file, indent=4)

    def setup(self, server, download_path, session_key, logfile=None, log_level=None, console_log_level=None):
        self.server = server
        self.download_path = Path(download_path).as_posix()
        self.session_key = session_key

        if not logfile:
            logfile = Path(self._get_dir(), 'gtagora-app-py.log')
        else:
            logfile = Path(logfile)
        self.logfile = str(logfile.as_posix())

        if not log_level:
            log_level = 'INFO'

        if not log_level in logging._nameToLevel:
            raise ValueError(f'Invalid log level. Possible levels are: CRITICAL, ERROR, WARNING, INFO, DEBUG')
        self.log_level = log_level

        if not console_log_level:
            console_log_level = 'INFO'

        if not console_log_level in logging._nameToLevel:
            raise ValueError(f'Invalid log level. Possible levels are: CRITICAL, ERROR, WARNING, INFO, DEBUG')
        self.console_log_level = console_log_level

        if not self.app_id:
            self.generate_app_id()
        self.save()

    def is_complete(self):
        for attr, value in self.__dict__.items():
            if value is None:
                return False

        return True


    def generate_app_id(self):
        self.app_id = str(uuid4())

    @staticmethod
    def get_path():
        return os.path.join(Settings._get_dir(), 'gt-agora-app-py.json')

    def _get(self):
        settings_file = self._get_or_create_file()
        with open(settings_file) as json_file:
            return json.load(json_file)

    def _get_or_create_file(self):
        if not os.path.exists(self._get_dir()):
            os.makedirs(self._get_dir(), exist_ok=True)

        if not os.path.exists(self.get_path()):
            with open(self.get_path(), "w") as file:
                file.write('{}')

        if not os.path.exists(self.get_path()):
            raise PermissionError('Cannot create the settings file')

        return self.get_path()

    @staticmethod
    def _get_dir():
        return os.path.expanduser('~') + '/.gyrotools/'



import json
import logging
import re
import subprocess
from base64 import b64decode
from logging.handlers import RotatingFileHandler
from pathlib import Path, PurePath
from typing import List

from gtagora import Agora
from gtagora.models.dataset import DatasetType

from gtagoraapp.details.settings import Settings
from gtagoraapp.details.ws import AgoraWebsocket


class App:
    def __init__(self):
        self.settings = Settings()

        if not self.settings.is_complete():
            raise ValueError('The app is not configured yet. Please run: "agoraapp.py --setup" first')

        self.logger = self._create_logger(self.settings.log_level)
        self.logger.info('Starting the Agora App...')

        self.agora = Agora.create(self.settings.server, token=self.settings.session_key)

        if not self.ping():
            raise ConnectionError('Cannot connect to Agora: The server could not be reached')

        if not self.check_connection():
            raise ConnectionError('Cannot connect to Agora: Please check your credentials')

        self.logger.info(f'Successfully connected to Agora at {self.settings.server}')

        self.websocket = AgoraWebsocket(self.settings, self)

    def run(self):
        self.websocket.run()

    def ping(self):
        return self.agora.ping()

    def check_connection(self):
        return self.agora.http_client.check_connection

    def download(self, files):
        try:
            self.logger.info(f'Downloading {len(files)} files:')
            count = 0
            for file in files:
                if len(file) == 5 or len(file) == 4:
                    id = file[0]
                    dir_name = file[1]
                    filename = file[2]
                    size = file[3]
                    if len(file) == 5:
                        hash = file[4]
                    else:
                        hash = None

                    download_path = PurePath(self.settings.download_path, dir_name)
                    Path(download_path).mkdir(parents=True, exist_ok=True)
                    download_file = PurePath(download_path, filename)
                    url = f'/api/v1/datafile/{id}/download/'
                    self.logger.info(f'    {self.agora.http_client.connection.url}/{url}  -->  {download_file.as_posix()}')
                    self.agora.http_client.download(url, download_file.as_posix())
                    count += 1
            self.logger.info(f'Successfully downloaded {count} files')
        except Exception as e:
            self.logger.error(f'Error downloading files: {str(e)}')

        self.logger.info(f'Download Complete')
        self.logger.info(f' ')

    def run_task(self, data):
        try:
            name = data.get('name')
            self.logger.info(' ')
            self.logger.info(f'Running task: "{name}"')

            task_info_id = data.get('taskInfo')
            output_directory = data.get('outputDirectory')
            commandline = data.get('commandLine')
            outputs = data.get('outputs')
            target = data.get('target')
            script = data.get('script')
            script_path = data.get('scriptPath')

            # download datafiles
            files = data.get('files')
            if files:
                self.download(files)

            # replace base path
            (output_directory, replacements) = self._prepare_output_directory(output_directory)

            if script and script_path:
                if not self._save_script(script_path, script):
                    return

            stdout = None
            error = None
            if commandline:
                self.logger.debug(f'  Replacing placeholders in command line:')
                commandline = self._replace_base_path(commandline, replacements)

                (data, error, stdout) = self._perform_task(commandline)

            if outputs and output_directory and target:
                files = self._collect_outputs(outputs, output_directory)

                if files:
                    target_id = target[0]
                    target_type = target[1]

                    self._upload_files(target_id, target_type, files)
                else:
                    self.logger.debug(f'  No files found to upload')

            if task_info_id:
                self._mark_task_as_finished(task_info_id, data, error)
                if stdout:
                    self._upload_stdout(task_info_id, stdout)

        except Exception as e:
            self.logger.error(f'Error executing task: {str(e)}')

        self.logger.info(f'Task Complete')
        self.logger.info(f' ')

    def _perform_task(self, command):
        data = dict()
        error = None
        self.logger.info(f'  Executing command:')
        self.logger.info(f'    {command}')
        data['command'] = command
        try:
            stdout = subprocess.check_output(command, stderr=subprocess.STDOUT)
            data['exit_code'] = 0
            self.logger.info(f'  Command successfully completed')
        except subprocess.CalledProcessError as e:
            error = f'{e.output}'
            data['exit_code'] = e.returncode
            self.logger.error(f'The process returned a non-zero exit code: exit code: {e.returncode}; message: {e.output}')
        return (data, error, stdout)

    def _prepare_output_directory(self, output_directory):
        output_directory_orig = output_directory
        output_directory = self._replace_base_path(output_directory, [('{{BASE_PATH}}', self.settings.download_path)])
        output_directory = Path(output_directory)
        self.logger.debug(f'  Replaced placeholders in output path:')
        self.logger.debug(f'    {output_directory_orig} -->  {output_directory_orig}')
        output_directory.mkdir(parents=True, exist_ok=True)
        output_directory_orig_data = (Path(output_directory_orig).parent / 'data').as_posix()
        output_directory_data = output_directory.parent / 'data'
        replacements = [(output_directory_orig, str(output_directory)),
                        (output_directory_orig_data, str(output_directory_data))]
        return (output_directory, replacements)

    def _replace_base_path(self, input_str: str, replacements: List[tuple]):
        output_str = input_str
        for replacement in replacements:
            output_str = output_str.replace(replacement[0], replacement[1])
            self.logger.debug(f'    {replacement[0]} -->  {replacement[1]}')
        return output_str

    def _save_script(self, script_path, script):
        script_path = script_path.replace('{{BASE_PATH}}', self.settings.download_path)
        self.logger.debug(f'  Saving the script to {script_path}')
        scriptDir = Path(script_path).parent
        scriptDir.mkdir(parents=True, exist_ok=True)
        decoded_script = b64decode(script)
        with open(script_path, 'wb') as file:
            file.write(decoded_script)
        if not Path(script_path).exists():
            self.logger.error(f'Cannot create the script to run')
            return False
        return True

    def _collect_outputs(self, outputs, output_directory):
        self.logger.info(f'  Collecting outputs:')
        files = []
        counter = 1
        for output in outputs:
            type = output.get('type')
            regex = output.get('regex')
            datasetType = output.get('datasetType')

            self.logger.info(f'    {counter}. type = {type}, regex = {regex}, datasetType = {datasetType}: ')

            if datasetType == DatasetType.PHILIPS_REC:
                cur_files = self._find_all_par_recs_in_directory(output_directory)
            elif datasetType == DatasetType.DICOM:
                cur_files = self._find_all_dicoms_in_directory(output_directory)
            elif datasetType == DatasetType.OTHER:
                cur_files = self._find_all_files_in_directory(output_directory)
            else:
                self.logger.error(f'The upload dataset type is not yet implemented. Please contact GyroTools')

            for file in cur_files:
                self.logger.info(f'      {file}')

            if not cur_files:
                self.logger.info(f'      No files found')

            files.extend(cur_files)
            counter += 1
        return files

    def _upload_files(self, target_id, target_type, files):
        try:
            self.logger.info(f'  Uploading output files to {target_type} {target_id}:')
            if target_type == 'folder':
                folder = self.agora.get_folder(target_id)
                folder.upload(files)
            elif target_type == 'serie':
                series = self.agora.get_series(target_id)
                series.upload(files)
            else:
                self.logger.error(f'Upload target type is not implemented yet. Please contact GyroTools')
        except Exception as e:
            self.logger.error(f'Error uploading files: {str(e)}')
            return False

    def _find_all_par_recs_in_directory(self, dir: Path, regex=None):
        if regex:
            r = re.compile(regex)
        else:
            r = re.compile('.*')

        recs = [f for f in dir.rglob('*.rec') if f.is_file() and r.match(f.as_posix())]
        pars = [f for f in dir.rglob('*.par') if f.is_file() and r.match(f.as_posix())]

        files = recs
        files.extend(pars)
        return files

    def _find_all_dicoms_in_directory(self, dir: Path, regex=None):
        if regex:
            r = re.compile(regex)
        else:
            r = re.compile('.*')

        return [f for f in dir.rglob('*') if f.is_file() and self.is_dicom_file(f) and r.match(f.as_posix())]

    def _find_all_files_in_directory(self, dir: Path, regex=None):
        if regex:
            r = re.compile(regex)
        else:
            r = re.compile('.*')

        return [f for f in dir.rglob('*') if f.is_file() and r.match(f.as_posix())]

    def _mark_task_as_finished(self, timeline_id, data, error):
        url = f'/api/v2/timeline/{timeline_id}/finish/'
        self.logger.debug(f'  Mark task as finished with url: {url}')
        status = self.agora.http_client.post(url, data={'data': json.dumps(data), 'error': error})
        if status.status_code == 404:
            url = f'/api/v1/taskinfo/{timeline_id}/finish/'
            self.logger.debug(f'  Mark task as finished with url: {url}')
            status = self.agora.http_client.post(url, data={'data': json.dumps(data), 'error': error})
            if status.status_code != 200:
                self.logger.warning(f'Could not mark the task as finish. status = {status.status_code}')

    def _upload_stdout(self, timeline_id, stdout):
        url = f'/api/v2/timeline/{timeline_id}/stdout/'
        self.logger.debug(f'  Send stdout to url: {url}')
        status = self.agora.http_client.post(url, data=stdout)
        if status.status_code == 404:
            url = f'/api/v1/taskinfo/{timeline_id}/stdout/'
            self.logger.debug(f'  Send stdout to url: {url}')
            status = self.agora.http_client.post(url, data=stdout)
            if status.status_code != 200:
                self.logger.warning(f'Could not upload the stdout. status = {status.status_code}')

    def _create_logger(self, level='INFO'):
        rotating_logger = logging.getLogger('gtagora-app-py')
        rotating_logger.setLevel(level)

        handler = RotatingFileHandler(
            self.settings.logfile, maxBytes=1024*1024*10, backupCount=10)
        formatter = logging.Formatter(fmt='%(asctime)s %(levelname)s:  %(message)s', datefmt='%m/%d/%Y %H:%M:%S')
        handler.setFormatter(formatter)
        rotating_logger.addHandler(handler)

        console_handler = logging.StreamHandler()
        console_formatter = logging.Formatter(fmt='%(levelname)s:  %(message)s')
        console_handler.setFormatter(console_formatter)
        console_handler.setLevel(logging.INFO)
        rotating_logger.addHandler(console_handler)
        return rotating_logger

    @staticmethod
    def is_dicom_file(file: Path):
        try:
            with file.open("rb") as f:
                f.seek(128)
                magic = f.read(4)
                return magic == b'DICM'
        except:
            logging.getLogger('gtagora-app-py').debug(f'Dicom check failed on file {str(file)}')
            return False

    @staticmethod
    def get_session_key(server:str, username: str, password: str):
        logging.getLogger('gtagora-app-py').info(f'Getting session key for user {username}')
        A = Agora.create(server, user=username, password=password)
        if not A.ping():
            raise ConnectionError('Cannot connect to Agora: The server could not be reached')

        if not A.http_client.check_connection():
            raise ConnectionError('Cannot connect to Agora: Please check your credentials')

        return A.http_client.connection.token



